;; -*- no-byte-compile: t; lexical-binding: nil -*-
(define-package "bui" "20260502.730"
  "Buffer interface library."
  '((emacs "29.1"))
  :url "https://github.com/alezost/bui.el"
  :commit "4319e1bf3ff94ff0568eed280ac1f980a4b68679"
  :revdesc "4319e1bf3ff9"
  :keywords '("tools")
  :authors '(("Alex Kost" . "alezost@gmail.com"))
  :maintainers '(("Alex Kost" . "alezost@gmail.com")))
